from typing import Dict

from django.http import Http404
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import ensure_csrf_cookie

from hx_requests.hx_registry import HXRequestRegistry
from hx_requests.utils import deserialize_kwargs, is_htmx_request


@method_decorator(ensure_csrf_cookie, name="dispatch")
class HtmxViewMixin:
    """
    Mixin to be added to views that are using HXRequests.
    Hijacks the get and post to route them to the proper
    HXRequest.
    """

    hx_requests: Dict = []

    def dispatch(self, request, *args, **kwargs):
        # Try to dispatch to the right method; if a method doesn't exist,
        # defer to the error handler. Also defer to the error handler if the
        # request method isn't on the approved list.
        if request.method.lower() in self.http_method_names:
            handler_class = self
            # If it's an HTMX request, use the HXRequest class to handle the request
            # otherwise, use the view class to handle the request.
            if is_htmx_request(request):
                kwargs.update(self.get_hx_extra_kwargs(request))
                handler_class = self._setup_hx_request(request, *args, **kwargs)
            handler = getattr(handler_class, request.method.lower(), self.http_method_not_allowed)
        else:
            handler = self.http_method_not_allowed
        return handler(request, *args, **kwargs)

    def get_hx_request(self, request):
        hx_request_name = request.GET.get("hx_request_name")
        hx_request_class = HXRequestRegistry.get_hx_request(hx_request_name)
        if not hx_request_class:
            raise Http404(
                f"No HXRequest found with the name {hx_request_name}. Are you sure it's spelled correctly?"
            )
        return hx_request_class()

    def get_hx_extra_kwargs(self, request):
        kwargs = {}
        for key in request.GET:
            kwargs[key] = request.GET.get(key)

        return deserialize_kwargs(**kwargs)

    def _setup_hx_request(self, request, *args, **kwargs):
        hx_request = self.get_hx_request(request)
        hx_request.view = self
        hx_request._setup_hx_request(request, *args, **kwargs)
        return hx_request



import json
from urllib.parse import quote_plus, urlencode, urlparse

from django.apps import apps
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models

MODEL_INSTANCE_PREFIX = "model_instance__"
KWARG_PREFIX = "___"
__ = "__"


def serialize(value):
    if isinstance(value, models.Model):
        return (
            f"{MODEL_INSTANCE_PREFIX}{value._meta.app_label}{__}{value._meta.model_name}{__}{value.pk}"
        )
    return json.dumps(value, cls=DjangoJSONEncoder)


def deserialize(value):
    if value.startswith(MODEL_INSTANCE_PREFIX):
        value = value.replace(MODEL_INSTANCE_PREFIX, "")
        app_label, model_name, pk = value.split(__)
        model = apps.get_model(app_label, model_name)
        return model.objects.get(pk=pk)
    return json.loads(value)


def is_htmx_request(request):
    return "HX-Request" in request.headers


def serialize_kwargs(**kwargs):
    serialized_kwargs = {}
    for k, v in kwargs.items():
        prefixed_k = KWARG_PREFIX + k
        serialized_kwargs[prefixed_k] = serialize(v)
    return serialized_kwargs


def deserialize_kwargs(**kwargs):
    deserialized_kwargs = {}
    for k, v in kwargs.items():
        if not k.startswith(KWARG_PREFIX):
            continue
        prefixed_k = k.replace(KWARG_PREFIX, "")
        deserialized_kwargs[prefixed_k] = deserialize(v)
    return deserialized_kwargs


def get_url(context, hx_request_name, obj, use_full_path=False, **kwargs):
    request = context["request"]
    url = request.path

    params = {"hx_request_name": hx_request_name}

    if use_full_path:
        get_params = {}
        for k, v in request.GET.lists():
            if k not in ["hx_request_name", "object"] and not k.startswith(KWARG_PREFIX):
                get_params[k] = v[0] if len(v) == 1 else v
        params.update(get_params)

    if obj:
        params["object"] = serialize(obj)

    # Parse the URL to check if there are existing query parameters
    parsed_url = urlparse(url)
    if parsed_url.query:  # If there are existing query parameters
        url += "&"  # Append using "&"
    else:
        url += "?"  # Otherwise, start with "?"

    url += urlencode(params, doseq=True)

    serialized_kwargs = serialize_kwargs(**kwargs)
    url += "".join(f"&{k}={quote_plus(str(v))}" for k, v in serialized_kwargs.items())

    return url


def get_csrf_token(context):
    cookie = context["request"].headers.get("cookie")
    if cookie and "csrftoken" in cookie:
        token = cookie.split("csrftoken=")[1].split(";")[0]
        return token



from django.http import Http404
from django.utils.translation import gettext_lazy as _


def list_view_get(self, request, *args, **kwargs):
    self.object_list = self.get_queryset()
    allow_empty = self.get_allow_empty()

    if not allow_empty:
        # When pagination is enabled and object_list is a queryset,
        # it's better to do a cheap query than to load the unpaginated
        # queryset in memory.
        if self.get_paginate_by(self.object_list) is not None and hasattr(self.object_list, "exists"):
            is_empty = not self.object_list.exists()
        else:
            is_empty = not self.object_list
        if is_empty:
            raise Http404(
                _("Empty list and “%(class_name)s.allow_empty” is False.")
                % {
                    "class_name": self.__class__.__name__,
                }
            )


def update_view_get(self, request, *args, **kwargs):
    self.object = self.get_object()


def create_view_get(self, request, *args, **kwargs):
    self.object = None


def detail_view_get(self, request, *args, **kwargs):
    self.object = self.get_object()


def delete_view_get(self, request, *args, **kwargs):
    self.object = self.get_object()


def date_view_get(self, request, *args, **kwargs):
    self.date_list, self.object_list, extra_context = self.get_dated_items()
    return {
        "object_list": self.object_list,
        "date_list": self.date_list,
        **extra_context,
    }



import importlib
import inspect
import threading

from django.apps import apps

from hx_requests.hx_requests import BaseHXRequest


class HXRequestRegistry:
    _registry = {}
    _initialized = False
    _lock = threading.Lock()  # Lock to ensure thread safety
    _processed_classes = set()  # Keep track of processed classes

    @classmethod
    def initialize(cls):
        if cls._initialized:
            return

        with cls._lock:  # Ensure that only one thread can initialize at a time
            if cls._initialized:
                return  # Another thread might have finished initializing while we were waiting

            # Proceed with initialization
            for app in apps.get_app_configs():
                try:
                    module = importlib.import_module(f"{app.label}.hx_requests")
                    clsmembers = inspect.getmembers(module, inspect.isclass)
                    for _, obj in clsmembers:
                        # Check if the class has already been processed
                        if obj in cls._processed_classes:
                            continue
                        cls._processed_classes.add(obj)

                        if issubclass(obj, BaseHXRequest) and getattr(obj, "name", None):
                            cls.register_hx_request(obj.name, obj)
                except ModuleNotFoundError:
                    continue

            cls._initialized = True

    @classmethod
    def register_hx_request(cls, name, hx_request_class):
        if name in cls._registry:
            raise Exception(f"Duplicate HXRequest name found: {name}")
        cls._registry[name] = hx_request_class

    @classmethod
    def get_hx_request(cls, name):
        cls.initialize()
        return cls._registry.get(name)

    @classmethod
    def get_all_hx_requests(cls):
        cls.initialize()
        return cls._registry



from typing import Dict

from django import template

from hx_requests.utils import get_csrf_token, get_url

register = template.Library()


@register.simple_tag(takes_context=True)
def hx_get(context: Dict, hx_request_name: str, object=None, use_full_path=False, **kwargs) -> str:
    """
    Renders the proper HTML attributes for an HX GET request.
    And includes the hx_request_name, object, and any additional kwargs.
    """
    url = get_url(context, hx_request_name, object, use_full_path, **kwargs)
    return f"hx-get={url}"


@register.simple_tag(takes_context=True)
def hx_post(context: Dict, hx_request_name: str, object=None, use_full_path=False, **kwargs) -> str:
    """
    Renders the proper HTML attributes for an HX POST request.
    And includes the hx_request_name, object, and any additional kwargs.
    Also includes the CSRF token if available.
    """
    url = get_url(context, hx_request_name, object, use_full_path, **kwargs)
    hx_attrs = f"hx-post={url}"
    token = get_csrf_token(context)
    if token:
        hx_attrs += f' hx-headers={{"X-CSRFTOKEN":"{token}"}}'
    return hx_attrs


@register.simple_tag(takes_context=True)
def hx_url(context: Dict, hx_request_name: str, object=None, use_full_path=False, **kwargs) -> str:
    """
    Returns the URL for an HX request. Can be used when you need to manually set the URL in a template.
    Useful if the hx-get or hx-post tags are not flexible enough for your use case.
    For example:

    ```
    hx-get="{% hx_url 'my-hx-request' object %}"

    ```
    """
    return get_url(context, hx_request_name, object, use_full_path, **kwargs)
